unit NGpuInfo;

{$IMPORTEDDATA OFF}

interface

uses NTypes;

type
  TGpuInfo = record
    DeviceName: WideString;
    RunningTime: Int64; // nanosec
    Temperature: Integer;
  end;

  TGpuInfoArray = array of TGpuInfo;

function GpuInfo: TGpuInfoArray;

implementation

uses Windows, D3DKmtSub, SetupapiSub;

var
  GpuInfoItems: TGpuInfoArray;

procedure InitGpuInfo;
var
  devinfo: HDEVINFO;
  data: SP_DEVINFO_DATA;
  id: array [0..MAX_DEVICE_ID_LEN] of WideChar;
  size: DWORD;
  i, j: Integer;
begin
  if not Assigned(D3DKMTOpenAdapterFromDeviceName) then
    Exit;
  devinfo := SetupDiGetClassDevsW(GUID_DISPLAY_DEVICE_ARRIVAL, nil, 0,
      DIGCF_PRESENT or DIGCF_DEVICEINTERFACE);
  if devinfo = INVALID_HANDLE_VALUE then
    Exit;
  data.cbSize := SizeOf(data);
  i := 0;
  while SetupDiEnumDeviceInfo(devinfo, i, @data) do begin
    if not SetupDiGetDeviceInstanceIdW(devinfo, @data, @id,
        SizeOf(id) div SizeOf(WideChar), size) then
      Exit;
    if Copy(id, 1, 5) <> 'ROOT\' then begin
      for j := 0 to size - 1 do
        if id[j] = '\' then id[j] := '#';
      SetLength(GpuInfoItems, Length(GpuInfoItems) + 1);
      GpuInfoItems[High(GpuInfoItems)].DeviceName :=
          '\\?\' + id + '#{1ca05180-a699-450a-9a0c-de4fbe3ddd89}';
    end;
    Inc(i);
  end;
end;

function GpuInfo: TGpuInfoArray;
var
  i: Integer;
  oa: D3DKMT_OPENADAPTERFROMDEVICENAME;
  qa: D3DKMT_QUERYADAPTERINFO;
  pd: D3DKMT_ADAPTER_PERFDATA;
  qs: D3DKMT_QUERYSTATISTICS;
  ca: D3DKMT_CLOSEADAPTER;
begin
  qa.Type_ := KMTQAITYPE_ADAPTERPERFDATA;
  qa.pPrivateDriverData := @pd;
  qa.PrivateDriverDataSize := SizeOf(pd);
  for i := 0 to High(GpuInfoItems) do
    with GpuInfoItems[i] do begin
      oa.pDeviceName := PWideChar(DeviceName);
      if D3DKMTOpenAdapterFromDeviceName(oa) <> 0 then
        Continue;

      qa.hAdapter := oa.hAdapter;
      pd.PhysicalAdapterIndex := i;
      if D3DKMTQueryAdapterInfo(qa) = 0 then
        if pd.Temperature > 0 then
          Temperature := (pd.Temperature + 5) div 10
        else
          Temperature := (pd.Temperature - 5) div 10;

      FillChar(qs, SizeOf(qs), 0);
      qs.AdapterLuid := oa.AdapterLuid;
      qs.Type_ := D3DKMT_QUERYSTATISTICS_NODE;
      if D3DKMTQueryStatistics(qs) = 0 then
        with qs.QueryResult.NodeInformation do
          RunningTime := GlobalInformation.RunningTime.QuadPart * 100;

      ca.hAdapter := oa.hAdapter;
      D3DKMTCloseAdapter(ca);
    end;
  Result := GpuInfoItems;
end;

begin
  InitGpuInfo;
end.
